"""
Test cases for other utilities
Created on: 27 August 2021.
"""
import albumentations
import jpeg4py as jpeg
import glob
import numpy as np 
import pandas as pd 
import os 
from pathlib import Path
from PIL import Image
from sklearn.model_selection import StratifiedKFold
from sklearn.metrics import roc_auc_score, f1_score
from tqdm.auto import tqdm

import torch
import torch.nn as nn 
import torch.utils.data as D 
from torchvision import models

import pytest
from tpu_util.other_utils import *

# Setup for the whole test case process
# Code taken from https://www.kaggle.com/wabinab/submission-gnet
# and another code taken from a private notebook https://www.kaggle.com/wabinab/gnet-tpu-train
# and another private notebook https://www.kaggle.com/wabinab/gnet-create-ds
FLAGS = {
    "data_dir": Path("./sample"),
    "num_workers": os.cpu_count(),
    "lr": 0.001,
    "bs": 8,
    "num_epochs": 2,
    "seed": 1447,
}


class GnetTestDataset(D.Dataset):
    def __init__(self, parent_path, df=None, shuffle=False, seed=None, 
                transforms=None, split=None):
        """
        parent_path: (pathlib.Path) Parent path of images. 
        df: (numpy) (This haven't changed the 'df' name, but is a numpy array of test files)
        shuffle: (Boolean) Shuffle dataset? Default: False. 
        seed: (integer) seed. Default: None. 
        transforms: (Albumentations) Transformation to images, default: None. 
        split: (python list) Splits index for train test split. 
        """
        self.df_np = df
        self.parent_path = parent_path
        self.seed = seed
        self.transforms = transforms

        if seed:
            np.random.seed(seed)
            torch.manual_seed(seed)

        if type(split) != type(None): self.df_np = self.df_np[split]

    def __len__(self): 
        return len(self.df_np)

    def __getitem__(self, idx):
        # item_path = self.parent_path / self.df_np[idx]  # image filenames
        item_path = self.df_np[idx]  # this requires changing depending on different cases. 
        image = jpeg.JPEG(item_path).decode()

        if self.transforms is not None:
            image = self.transforms(**{"image": image})
            image = image["image"]

        image = torch.from_numpy(image).permute(2, 0, 1)
        
        return torch.Tensor([idx]).type(torch.uint8), image


class GnetDataset(D.Dataset):
    def __init__(self, parent_path, df=None, shuffle=False,
                 seed=None, transforms=None, split=None):
        """
        parent_path: (pathlib.Path) Parent path of images. 
        df: (pandas.DataFrame) Dataframe containing labels and file path. 
        shuffle: (Boolean) Shuffle dataset? Default: False. 
        seed: (integer) seed. Default: None. 
        transforms: (Albumentations) Transformation to images, default: None. 
        split: (python list) Splits index for train test split. 
        """
        self.df_np = df.to_numpy()
        self.parent_path = parent_path
        self.seed = seed
        self.transforms = transforms
        
        if seed:
            np.random.seed(seed)
            torch.manual_seed(seed)
        
        if type(split) != type(None): self.df_np = self.df_np[split]
            
    def __len__(self): 
        return len(self.df_np)
    
    def __getitem__(self, idx):
        item = self.df_np[idx]
        item_path = self.parent_path / item[0]  # image filenames
        image = jpeg.JPEG(item_path).decode()
        
        if self.transforms is not None:
            image = self.transforms(**{"image": image})
            image = image["image"]
            
        image = torch.from_numpy(image).permute(2, 0, 1)  # HWC -> CHW format
        target = torch.Tensor([item[1]])
        
        return image, target


test_path = np.array([Path(path) for path in sorted(glob.glob("./sample/*.jpg"))])
test_ds = GnetTestDataset(FLAGS["data_dir"], test_path, seed=FLAGS["seed"])

df = pd.read_csv("./sample/training_labels.csv")
df["id"] = df["id"] + ".jpg"

skf = StratifiedKFold(n_splits=5)
tdf_np = df.to_numpy()
X = tdf_np[:, 0]
y = tdf_np[:, 1].astype(np.uint8)
for train_index, test_index in skf.split(X, y): pass

def get_dataset():
    parent_path = FLAGS["data_dir"]

    train_ds = GnetDataset(parent_path, df, shuffle=True, seed=FLAGS["seed"],
                            split=train_index)
    val_ds = GnetDataset(parent_path, df, shuffle=False, seed=FLAGS["seed"],
                            split=test_index)

    return train_ds, val_ds


model = models.resnet50(pretrained=True)
num_features = model.fc.in_features
for param in model.parameters(): param.requires_grad_(False)
model.fc = nn.Linear(num_features, 1)  
# actually requires sigmoid but we're not using model here so ignore. 

criterion = nn.BCEWithLogitsLoss()
opt = torch.optim.Adam(model.parameters(), lr=FLAGS["lr"])


### ---------------------------------------------------- ###
### Starting test case ###
### ---------------------------------------------------- ###
@pytest.fixture
def call_dl(tmpdir):
    train_ds, val_ds = get_dataset()
    return dataloader(train_ds, val_ds, FLAGS, test_ds)


def test_dataloader_returns_correct_data_type():
    """For distributed=False. Unfortunately we cannot test distributed=True without TPU."""
    train_ds, val_ds = get_dataset()
    dls = dataloader(train_ds, val_ds, FLAGS)

    assert type(dls) == dict


def test_dataloader_have_correct_items(call_dl):
    """Should have train, val, and optionally test"""

    assert "train" in call_dl.keys()
    assert "val" in call_dl.keys()
    assert "test" in call_dl.keys()


def test_dataloader_each_child_item_correct_datatype(call_dl):
    for key in call_dl.keys():
        assert type(call_dl[key]) == torch.utils.data.dataloader.DataLoader


def test_dataloader_no_exception_raise_calling_one_batch_train(call_dl):
    try: 
        for one_batch in call_dl["train"]: break
    except Exception: pytest.fail("cannot call one batch")


def test_dataloader_no_exception_raise_calling_one_batch_val(call_dl):
    try:
        for one_batch in call_dl["val"]: break
    except Exception: pytest.fail("cannot call one batch")


def test_dataloader_no_exception_raise_calling_one_batch_test(call_dl):
    try:
        for one_batch in call_dl["test"]: break
    except Exception: pytest.fail("cannot call one batch")


def test_linear_scheduler_correct_value():
    steps = 10
    lin_sched = LinearScheduler(opt, 10, steps)
    all_lr = np.zeros((steps, ))

    for k, i in enumerate(range(steps)):
        all_lr[k] = lin_sched.get_lr()[0]
        lin_sched.step()

    assert (np.round(all_lr).astype(np.uint8) == np.arange(10) + 1).all()


def test_exponential_scheduler_correct_value():
    steps, end_lr = 10, 10
    exp_sched = ExponentialScheduler(opt, end_lr, steps)
    all_lr = np.zeros((steps, ))

    comparable = np.array([FLAGS["lr"] * ((end_lr / FLAGS["lr"]) ** (pct / steps)) for pct in range(1, steps + 1)])

    for k, i in enumerate(range(steps)):
        all_lr[k] = exp_sched.get_lr()[0]
        exp_sched.step()

    assert (all_lr == comparable).all()


def test_cosine_scheduler_correct_value():
    steps, end_lr = 10, 10
    cos_sched = CosineScheduler(opt, end_lr, steps)
    all_lr = np.zeros((steps, ))

    comparable = np.array([end_lr + (FLAGS["lr"] - end_lr) / 2 * (np.cos(np.pi * (pct / steps)) + 1) for pct in range(1, steps + 1)])

    for k, i in enumerate(range(steps)):
        all_lr[k] = cos_sched.get_lr()[0]
        cos_sched.step()

    assert (all_lr == comparable).all()


def test_annealing_no_functions_correctly():
    assert annealing_no(1, 10, 0.1) == 1
    assert annealing_no(2, 10, 0.1) == 2
    assert annealing_no(1, 2, 0.1) == 1
    assert annealing_no(1, 10, 0.5) == 1


@pytest.fixture
def start_pct_end(tmpdir): return 0.001, 0.2, 10


def test_annealing_linear_functions_correctly(start_pct_end):
    start, pct, end = start_pct_end
    assert annealing_linear(start, end, pct) == pytest.approx(2, 0.01)
    assert annealing_linear(1, end, pct) == 2.8
    assert annealing_linear(start, 100, pct) == pytest.approx(20, 0.001)
    assert annealing_linear(start, end, 0.5) == pytest.approx(5, 0.01)


def test_annealing_exp_functions_correctly(start_pct_end):
    start, pct, end = start_pct_end
    assert annealing_exp(start, end, pct) == start * ((end / start) ** pct)
    assert annealing_exp(1, end, pct) == 1 * ((end / 1) ** pct)
    assert annealing_exp(start, 100, pct) == start * ((100 / start) ** pct)
    assert annealing_exp(start, end, 0.5) == start * ((end / start) ** 0.5)


def test_annealing_cos_functions_correctly(start_pct_end):
    start, pct, end = start_pct_end
    assert annealing_cos(start, end, pct) == end + (((start - end) / 2) * (1 + np.cos(np.pi * pct)))
    assert annealing_cos(1, end, pct) == end + (((1 - end) / 2) * (1 + np.cos(np.pi * pct)))
    assert annealing_cos(start, 100, pct) == 100 + (((start - 100) / 2) * (1 + np.cos(np.pi * pct)))
    assert annealing_cos(start, end, 0.5) == end + (((start - end) / 2) * (1 + np.cos(np.pi * 0.5)))


# normalize_fn
def test_normalize_fn_default_correct_value():
    mean, std = normalize_fn()

    assert (mean.squeeze() == torch.Tensor([123.6750, 116.2800, 103.5300])).all()
    assert (std.squeeze() == torch.Tensor([58.3950, 57.1200, 57.3750])).all()


def test_normalize_fn_default_correct_shape():
    mean, std = normalize_fn()

    assert mean.shape == torch.Size([1, 3, 1, 1])
    assert std.shape == torch.Size([1, 3, 1, 1])


def test_normalize_fn_with_data(call_dl):
    for data, target in call_dl["train"]: break
    data_out = normalize_fn(data)
    mean, std = normalize_fn()

    assert (data_out == (data - mean) / std).all()


def test_normalize_fn_with_data_shape_retains(call_dl):
    for data, target in call_dl["train"]: break
    data_out = normalize_fn(data)

    assert data_out.shape == data.shape


def test_normalize_fn_calculated_input_false(call_dl):
    for data, target in call_dl["train"]: break
    mean, std = normalize_fn()

    data_out = normalize_fn(data, mean, std, calculated_input=True)
    data_baseline = normalize_fn(data)

    assert (data_out == data_baseline).all()


def test_normalize_fn_float_mean_works():
    mean, std = normalize_fn(mean=0.495)

    one_mean = 0.495 * 255
    assert (mean.squeeze() == torch.Tensor([one_mean, one_mean, one_mean])).all()


def test_normalize_fn_float_std_works():
    mean, std = normalize_fn(std=0.38)
    one_std = 0.38 * 255

    assert (std.squeeze() == torch.Tensor([one_std, one_std, one_std])).all()


def test_normalize_fn_int_mean_works():
    mean, std = normalize_fn(mean=3)

    one_mean = 3 * 255
    assert (mean.squeeze() == torch.Tensor([one_mean, one_mean, one_mean])).all()


def test_normalize_fn_int_std_works():
    mean, std = normalize_fn(std=3)

    one_std = 3 * 255
    assert (std.squeeze() == torch.Tensor([one_std, one_std, one_std])).all()


def test_normalize_fn_max_pixel_works():
    max_pixel = 2**16
    mean, std = normalize_fn(max_pixel=max_pixel)

    mean_comp = np.array([0.485, 0.456, 0.406], dtype=np.float32) * max_pixel
    std_comp = np.array([0.229, 0.224, 0.225], dtype=np.float32) * max_pixel

    assert mean.squeeze().numpy() == pytest.approx(mean_comp, 1e-6)
    assert std.squeeze().numpy() == pytest.approx(std_comp, 1e-6)


def test_max_pixel_can_be_float():
    mean, std = normalize_fn(max_pixel=0.496)
    compare = np.array([0.485, 0.456, 0.406], dtype=np.float32) * 0.496
    assert mean.squeeze().numpy() == pytest.approx(compare, 1e-6)


def test_normalize_fn_str_doesnt_work_mean(): 
    with pytest.raises(np.core._exceptions.UFuncTypeError):
        mean, std = normalize_fn(mean="0.64")


def test_normalize_fn_str_doesnt_work_std(): 
    with pytest.raises(np.core._exceptions.UFuncTypeError):
        mean, std = normalize_fn(std="0.64")


def test_normalize_fn_str_doesnt_work_max_pixel(): 
    with pytest.raises(np.core._exceptions.UFuncTypeError):
        mean, std = normalize_fn(max_pixel="0.64")


def test_normalize_fn_with_another_shape_mean_works():
    mean, std = normalize_fn(mean=(0.5, 0.6, 0.7, 0.8, 0.9))

    assert mean.shape == torch.Size([1, 5, 1, 1])





# %% [code]
"""
Contains utilities that does not requires TPU to run. 
Works together with tpu_utility_1.py which contains functions
requiring TPU to run. 
Separating them allow tests to be run. 
Created on: 27 August 2021.
"""
import copy
import numpy as np 
import pandas as pd 
import matplotlib.pyplot as plt
from IPython.display import clear_output

import torch
import torch.nn as nn
import torch.utils.data as D
from torch.optim import lr_scheduler

from pathlib import Path
from tqdm.auto import tqdm

# %% [code] {"jupyter":{"outputs_hidden":false},"execution":{"iopub.status.busy":"2021-08-25T05:58:03.379271Z","iopub.execute_input":"2021-08-25T05:58:03.379563Z","iopub.status.idle":"2021-08-25T05:58:03.388903Z","shell.execute_reply.started":"2021-08-25T05:58:03.379535Z","shell.execute_reply":"2021-08-25T05:58:03.387263Z"}}
def dataloader(train_ds, val_ds, flags, test_ds=None, distributed=False):
    """
    flags requirement: (python dict)
        "bs": (int) batch_size,
        "num_workers": (int) number of workers.

    :args have_test: (bool) return test dataloader? Default: False
        If True: requires test_ds to be defined. 
    """
    if distributed:
        import torch_xla.core.xla_model as xm
        train_sampler = D.distributed.DistributedSampler(
            train_ds, num_replicas=xm.xrt_world_size(),
            rank=xm.get_ordinal(), shuffle=True)
        train_loader = D.DataLoader(train_ds, batch_size=flags["bs"], sampler=train_sampler,
                                   num_workers=flags["num_workers"], drop_last=True)
    else:
        train_loader = D.DataLoader(train_ds, batch_size=flags["bs"], shuffle=True,
                                   num_workers=flags["num_workers"], drop_last=False)
        
    val_loader = D.DataLoader(val_ds, batch_size=flags["bs"], shuffle=False,
                             num_workers=flags["num_workers"],
                             drop_last=True if distributed else False)
    
    if not test_ds: test_loader = None
    else: test_loader = D.DataLoader(test_ds, batch_size=flags["bs"], shuffle=False, 
                                    num_workers=flags["num_workers"])
    
    return {"train": train_loader, "val": val_loader, "test": test_loader}


# %% [markdown]
# ## Show one batch

# %% [code] {"jupyter":{"outputs_hidden":false},"execution":{"iopub.status.busy":"2021-08-25T05:58:03.440365Z","iopub.execute_input":"2021-08-25T05:58:03.440666Z","iopub.status.idle":"2021-08-25T05:58:03.451694Z","shell.execute_reply.started":"2021-08-25T05:58:03.440634Z","shell.execute_reply":"2021-08-25T05:58:03.450924Z"}}
def show_one_batch(dls, nrows=2, ncols=4, aspect=None):
    """
    Show image of a single batch up to nrows * ncols images. 
    Will raise error if nrows * ncols is larger than the number of batches. 
    The title of each subplots are their labels (same as fastai). 
    
    Args:
        dls: (python dict) Containing keys "train" and "test" corresponding to 
            PyTorch DataLoader. Use `dataloader` function to get this. 
        nrows: (int) matplotlib subplot args. 
        ncols: (int) matplotlib subplot args. 
        aspect: (str/others) matplotlib imshow args. Default: None. 
    For matplotlib args, check matplotlib respective docs for what is supported. 
    """
    for data, target in dls["train"]: break
    
    count = 0
    fig, ax = plt.subplots(nrows=nrows, ncols=ncols)
    for row in range(nrows):
        for col in range(ncols):
            ax[row, col].imshow(data[count].permute(1, 2, 0), aspect=aspect)
            ax[row, col].axis("off")
            ax[row, col].title.set_text(target[count])
            count += 1


# %% [code] {"jupyter":{"outputs_hidden":false},"execution":{"iopub.status.busy":"2021-08-25T05:58:03.453140Z","iopub.execute_input":"2021-08-25T05:58:03.453784Z","iopub.status.idle":"2021-08-25T05:58:03.471632Z","shell.execute_reply.started":"2021-08-25T05:58:03.453717Z","shell.execute_reply":"2021-08-25T05:58:03.470572Z"}}
class LinearScheduler(lr_scheduler._LRScheduler):
    """
    Linearly increases lr between two boundaries over a number of iterations. 
    """
    def __init__(self, opt, end_lr, num_iter):  # original `optimizer`
        self.end_lr = end_lr
        self.num_iter = num_iter
        super(LinearScheduler, self).__init__(opt)
        
    def get_lr(self):
        """Formula: start + pct * (end - start)"""
        curr_iter = self.last_epoch + 1
        pct = curr_iter / self.num_iter  # ratio
        return [base_lr + pct * (self.end_lr - base_lr) for base_lr in self.base_lrs]
    
    
class ExponentialScheduler(lr_scheduler._LRScheduler):
    """
    Exponentially increases lr between two boundaries over a number of iterations. 
    """
    def __init__(self, opt, end_lr, num_iter, last_epoch = -1):
        self.end_lr = end_lr
        self.num_iter = num_iter
        super(ExponentialScheduler, self).__init__(opt, last_epoch=last_epoch) ## 
        
    def get_lr(self):
        curr_iter = self.last_epoch + 1
        pct = curr_iter / self.num_iter
        return [base_lr * (self.end_lr / base_lr) ** pct for base_lr in self.base_lrs]
    
    
class CosineScheduler(lr_scheduler._LRScheduler):
    """
    Cosine increases lr between two boundaries over a number of iterations. 
    """
    def __init__(self, opt, end_lr, num_iter, last_epoch = -1):
        self.end_lr = end_lr
        self.num_iter = num_iter
        super(CosineScheduler, self).__init__(opt, last_epoch=last_epoch) ## 
        
    def get_lr(self):
        curr_iter = self.last_epoch + 1
        pct = curr_iter / self.num_iter
        cos_out = np.cos(np.pi * pct) + 1
        return [self.end_lr + (base_lr - self.end_lr) / 2 * cos_out for base_lr in self.base_lrs]


# %% [code] {"execution":{"iopub.status.busy":"2021-08-13T07:53:08.882115Z","iopub.execute_input":"2021-08-13T07:53:08.882454Z","iopub.status.idle":"2021-08-13T07:53:08.888884Z","shell.execute_reply.started":"2021-08-13T07:53:08.882424Z","shell.execute_reply":"2021-08-13T07:53:08.887981Z"},"jupyter":{"outputs_hidden":false}}
def annealing_no(start, end, pct):
    """No annealing, always return 'start'."""
    return start


def annealing_linear(start, end, pct):
    """Linearly anneal from start to end as pct goes from 0.0 to 1.0."""
    return start + (pct * (end - start))


def annealing_exp(start, end, pct):
    """Exponentially anneal from start to end as pct goes from 0.0 to 1.0."""
    return start * (end / start) ** pct


def annealing_cos(start, end, pct):
    """Cosine anneal from start and end as pct goes from 0.0 to 1.0."""
    cos_out = np.cos(np.pi * pct) + 1
    return end + (start - end) / 2 * cos_out


# %% [markdown]
# `train_tpu` normalize function gotten from https://github.com/albumentations-team/albumentations/blob/300ee99386ad27f482387047dac4f6dddff11ac2/albumentations/augmentations/functional.py#L131

# %% [code] {"jupyter":{"outputs_hidden":false}}
def normalize_fn(data=None, mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225), 
              max_pixel=255, calculated_input=False):
    """
    Normalize image function. 
    
    :input requirement: (PyTorch Tensor) of shape (Batch, channel, height, width)
    
    :args:
        :data: (The input). If passed in, will not return mean and std. If None, return
            mean and std instead. 
        :mean: (int/tuple) If int, will be broadcasted in the channel dimension. 
            If tuple, must have same number of values as number of channels. 
            Mean of values. 
        :std: (int/tuple) Check mean for explanation. Standard deviation of values. 
        :max_pixel: (int/float) This is the max pixel values. Default: 255 (so image are from 0-255).
        :calculated input: (bool) Whether the input are already calculated, as in 
            they are tensors to be used directly with the correct shape. 
        
    :return: 
        (data != None) normalized data, same shape as input. 
        (data is None) mean, std 
    """
    if not calculated_input: 
        if type(mean) in [float, int]: mean = [mean]
        if type(std) in [float, int]: std = [std]

        mean = np.array(mean) * max_pixel
        std = np.array(std) * max_pixel

        mean = torch.from_numpy(mean).view(1, mean.shape[0], 1, 1).type(torch.float32)
        std = torch.from_numpy(std).view(1, std.shape[0], 1, 1).type(torch.float32)
        
    if data == None: return mean, std
    
    assert type(mean) == torch.Tensor
    assert type(std) == torch.Tensor
    assert data.shape[1] == mean.shape[1] == std.shape[1]
    return (data - mean.to(data.device)) / std.to(data.device)


# %% [code] {"jupyter":{"outputs_hidden":false},"execution":{"iopub.status.busy":"2021-08-25T05:56:43.862555Z","iopub.execute_input":"2021-08-25T05:56:43.863001Z","iopub.status.idle":"2021-08-25T05:56:43.875244Z","shell.execute_reply.started":"2021-08-25T05:56:43.862917Z","shell.execute_reply":"2021-08-25T05:56:43.874047Z"}}
def setup_kaggle():
    os.system("curl https://raw.githubusercontent.com/pytorch/xla/master/contrib/scripts/env-setup.py -o pytorch-xla-env-setup.py")
    print("Download complete step 1 of 2")
    os.system("python pytorch-xla-env-setup.py --version 1.7 --apt-packages libomp5 libopenblas-dev")
    print("Setup complete step 2 of 2")
    
    clear_output()

# %% [code] {"jupyter":{"outputs_hidden":false},"execution":{"iopub.status.busy":"2021-08-25T05:56:43.877682Z","iopub.execute_input":"2021-08-25T05:56:43.878075Z","iopub.status.idle":"2021-08-25T05:58:01.695632Z","shell.execute_reply.started":"2021-08-25T05:56:43.878040Z","shell.execute_reply":"2021-08-25T05:58:01.694297Z"}}
import os
from IPython.display import clear_output

try: import torch_xla
except Exception: setup_kaggle()

from .tpu_utility_1 import *
from .tpu_cache_ds_utils import *
from .other_utils import *
from ._lr_finder import *


import setuptools

setuptools.setup(
    name="tpu_util",
    version="0.1.3",
    url="https://github.com/Wabinab/TPU_utility",
    description="TPU utility with PyTorch XLA",
    long_description=open("README.md").read(),
    packages=setuptools.find_packages(),
    install_requires= ["pip", "packaging"],
    python_requires=">=3.6",
)



# %% [code]
from .tpu_utility_1 import *

import torch_xla  # as a decoration here. 
import torch_xla.core.xla_model as xm
import torch_xla.debug.metrics as met
import torch_xla.distributed.parallel_loader as pl
import torch_xla.distributed.xla_multiprocessing as xmp
import torch_xla.utils.utils as xu
import torch_xla.utils.cached_dataset as xcd

import torch
import torch.nn as nn
import torch.utils.data as D
from torch.optim import lr_scheduler

import numpy as np 
import pandas as pd 
import os
from IPython.display import clear_output

# %% [code] {"jupyter":{"outputs_hidden":false}}
def cache_dataset(train_ds=None, val_ds=None, get_dataset=None, compress=False):
    """
    train_ds: (PyTorch Dataset) training dataset. 
    val_ds: (PyTorch Dataset) validation dataset.
    get_dataset (Python Function) (only used if train_ds and val_ds are not passed.)
        Function that returns train_ds and val_ds as tuple. 
    compress: Whether to .tar.gz the output folder. Defaults: False. 
    kwargs: keyword arguments that are passed into get_dataset. 
    """
    if (train_ds is get_dataset is None) or (val_ds is get_dataset is None): 
        raise ValueError("Please pass in (train_ds AND val_ds) OR get_dataset. ")
    
    def _mp_fn(index, train_ds, val_ds):
        if get_dataset: train_ds, val_ds = get_dataset()
        train_ds = xcd.CachedDataset(train_ds, "./cached-train")
        val_ds = xcd.CachedDataset(val_ds, "./cached-val")
        print("Creating training dataset.")
        train_ds.warmup()
        print("Done\nCreating validationd dataset.")
        val_ds.warmup()
        print("Done")
        
    xmp.spawn(_mp_fn, args=(train_ds, val_ds,), start_method="fork", nprocs=1)
    
    clear_output()
    
    if compress:
        print("Compressing validation dataset (before training dataset)")
        os.system("tar czf cached-val.tar.gz ./cached-val")
        print("Deleting original folder (to save disk space)")
        os.system("rm -r ./cached-val")
        print("Compressed validation dataset")
        
        print("Compressing training dataset")
        os.system("tar czf cached-train.tar.gz ./cached-train")
        print("Deleting original folder to save disk space")
        os.system("rm -r ./cached-train")
        print("Compresed training dataset.")


# %% [code] {"jupyter":{"outputs_hidden":false},"execution":{"iopub.status.busy":"2021-08-25T05:58:01.697697Z","iopub.execute_input":"2021-08-25T05:58:01.698083Z","iopub.status.idle":"2021-08-25T05:58:03.377702Z","shell.execute_reply.started":"2021-08-25T05:58:01.698048Z","shell.execute_reply":"2021-08-25T05:58:03.376787Z"}}
import copy
import numpy as np 
import pandas as pd 
import matplotlib.pyplot as plt
from IPython.display import clear_output

import torch_xla  # as a decoration here. 
import torch_xla.core.xla_model as xm
import torch_xla.debug.metrics as met
import torch_xla.distributed.parallel_loader as pl
import torch_xla.distributed.xla_multiprocessing as xmp
import torch_xla.utils.utils as xu
import torch_xla.utils.cached_dataset as xcd

import torch
import torch.nn as nn
import torch.utils.data as D
from torch.optim import lr_scheduler

from pathlib import Path
from tqdm.auto import tqdm

from sklearn.metrics import f1_score

from .other_utils import *
from ._lr_finder import *

# %% [code] {"jupyter":{"outputs_hidden":false},"execution":{"iopub.status.busy":"2021-08-25T05:58:03.390186Z","iopub.execute_input":"2021-08-25T05:58:03.390486Z","iopub.status.idle":"2021-08-25T05:58:03.408775Z","shell.execute_reply.started":"2021-08-25T05:58:03.390457Z","shell.execute_reply":"2021-08-25T05:58:03.407563Z"}}
def distrib_dataloader(get_dataset, flags, cached=False):
    """
    Will run without test_ds available.
    get_dataset: (function) The function that returns train_ds, val_ds as tuple. 
    flags: (python dict): Requires: 
        "bs": (int) batch_size,
        "num_workers": (int) number of workers.
    cached: Not yet implemented. 

    To assign test_ds, you can do dls["test"] = test_ds later on. 
    """
    serial_exec = xmp.MpSerialExecutor()
    
    train_ds, val_ds = serial_exec.run(get_dataset)
    dls = dataloader(train_ds, val_ds, flags, distributed=True)
    return dls

# %% [code] {"jupyter":{"outputs_hidden":false},"execution":{"iopub.status.busy":"2021-08-25T05:58:03.410078Z","iopub.execute_input":"2021-08-25T05:58:03.410580Z","iopub.status.idle":"2021-08-25T05:58:03.421279Z","shell.execute_reply.started":"2021-08-25T05:58:03.410544Z","shell.execute_reply":"2021-08-25T05:58:03.420437Z"}}
def train_cycle_distrib(dls, flags, train_loop_fn, val_loop_fn, device=None):
    """
    dls: dataloaders, use distrib_dataloader() to get this. 
    flags: (python dict). Required:
        "num_epochs": number of epochs to train for. 
        "metrics_debug": whether to print metrics report of TPU. 
    train_loop_fn: (function) The training loop function.
    val_loop_fn: (function) The validation loop function. 
    device = device. Defaults: xm.xla_device()
    """
    for epoch in range(1, flags["num_epochs"] + 1):
        para_loader = pl.ParallelLoader(dls["train"], [device])
        train_loop_fn(para_loader.per_device_loader(device))
        clear_output(wait=True)
        xm.master_print(f"Finished training epoch {epoch}")

        para_loader = pl.ParallelLoader(dls["val"], [device])
        returned_val = val_loop_fn(para_loader.per_device_loader(device))
        if flags["metrics_debug"]: xm.master_print(met.metrics_report(), flush=True)
        
    return returned_val

# %% [code] {"jupyter":{"outputs_hidden":false},"execution":{"iopub.status.busy":"2021-08-25T05:58:03.422940Z","iopub.execute_input":"2021-08-25T05:58:03.423371Z","iopub.status.idle":"2021-08-25T05:58:03.439128Z","shell.execute_reply.started":"2021-08-25T05:58:03.423341Z","shell.execute_reply":"2021-08-25T05:58:03.438030Z"}}
def cached_dataset(cache_train_loc=None, cache_val_loc=None):
    """
    NOTE: This SHOULD BE CALLED inside the `_mp_fn` function of distributed. 
    Will fetch the cached dataset that is already preprocessed. 
    """
    if cache_train_loc is None: cache_train_loc = "./cache_train"
    if cache_val_loc is None: cache_val_loc = "./cache_val"
        
    train_ds = xcd.CachedDataset(None, cache_train_loc)
    val_ds = xcd.CachedDataset(None, cache_val_loc)
    
    return train_ds, val_ds

# %% [markdown]
# # LR Finder TPU

# Originally taken from https://nbviewer.jupyter.org/github/aman5319/Multi-Label/blob/master/Classify_scenes.ipynb
# Some information are gotten from https://github.com/davidtvs/pytorch-lr-finder/blob/master/torch_lr_finder/lr_finder.py

# %% [code] {"jupyter":{"outputs_hidden":false},"execution":{"iopub.status.busy":"2021-08-25T06:04:09.838292Z","iopub.execute_input":"2021-08-25T06:04:09.838826Z","iopub.status.idle":"2021-08-25T06:04:09.845127Z","shell.execute_reply.started":"2021-08-25T06:04:09.838793Z","shell.execute_reply":"2021-08-25T06:04:09.844020Z"}}
def lr_finder(model, opt, criterion, dls=None, train_loader=None, device=None):
    """
    Learning Rate Finder with default values. 
    model: your model.
    opt: torch.optim optimizers.
    criterion: torch.nn loss function. 
    dls: dataloaders. Check out `dataloader` function. Don't pass train_loader if this is used.
    train_loader: train_ds. PyTorch Dataset. Don't pass dls if this is used. 
    device: device.
    
    Return: steepest_point, model, best_loss
    """
    lrfinder = LRFinder(model, opt, criterion, device=device, barrier=True)
    if dls is train_loader is None: raise ValueError("One of dls or train_loader must be passed.")
    return lrfinder.lr_find(dls["train"] if dls else train_loader)

# %% [code] {"jupyter":{"outputs_hidden":false}}
def distrib_lr_finder(model, opt, criterion, dls=None, train_loader=None, device=None):
    """
    Learning Rate Finder with default values. 
    Will use distributed method to allow definition of `xm.xla_device()` in training
    without requiring restart of jupyter notebook. 
    
    :Args:
        model: your model.
        opt: torch.optim optimizers.
        criterion: torch.nn loss function. 
        dls: dataloaders. Check out `dataloader` function. Don't pass train_loader if this is used.
        train_loader: PyTorch DataLoader. Using this if dls not detected. 
        device: device.
        
    :return: (Nothing is returned)
    """
    SERIAL_EXEC = xmp.MpSerialExecutor()
    WRAPPED_MODEL = xmp.MpModelWrapper(model)

    def train_fn(lrfinder, train_loader):
        device = xm.xla_device()
        model = WRAPPED_MODEL.to(device)
        if not xm.is_master_ordinal(): xm.rendezvous("lrfind_only_once")
        model, opt = SERIAL_EXEC.run(lambda: lrfinder.lr_find(train_loader, model=model, device=device))
        if xm.is_master_ordinal(): xm.rendezvous("lrfind_only_once")
        return model, opt

    def _mp_fn(rank, model, opt, criterion, device):
        lrfinder = LRFinder(model, opt, criterion, device=device, barrier=False)
        if dls is train_loader is None: raise ValueError("One of dls or train_loader must be passed.")
        model, opt = train_fn(lrfinder, dls["train"] if dls else train_loader)
        if rank == 0: 
            torch.save({
                "model": model.cpu().state_dict(),
                "opt": opt.cpu().state_dict(),
            }, "/kaggle/working/temp.pth")
        
    xmp.spawn(_mp_fn, args=(WRAPPED_MODEL, opt, criterion, device,), start_method="fork", nprocs=8)

# %% [markdown]
# # One Cycle Policy
# Taken from https://nbviewer.jupyter.org/github/aman5319/Multi-Label/blob/master/Classify_scenes.ipynb

# %% [code] {"execution":{"iopub.status.busy":"2021-08-13T07:53:04.364632Z","iopub.execute_input":"2021-08-13T07:53:04.365059Z","iopub.status.idle":"2021-08-13T07:53:04.372010Z","shell.execute_reply.started":"2021-08-13T07:53:04.365022Z","shell.execute_reply":"2021-08-13T07:53:04.370905Z"},"jupyter":{"outputs_hidden":false}}
class Stepper():
    """Step through n_iter on a schedule defined by func."""
    def __init__(self, val, n_iter: int, func):
        self.start, self.end = val
        self.n_iter = max(1, n_iter)
        self.func = func
        self.n = 0
        
    def step(self):
        """Returned next value along annealed schedule."""
        self.n += 1
        return self.func(self.start, self.end, self.n/self.n_iter)
    
    @property
    def is_done(self):
        """Return True if schedule complted."""
        return self.n >= self.n_iter
        

# %% [code] {"jupyter":{"outputs_hidden":false}}
class OneCyclePolicy_TPU:
    def __init__(self, model, opt, criterion, FLAGS, num_iter=None, train_ds=None,
                 momentum=(0.95, 0.85), div_factor=25, pct_start=0.4,
                # train_transform=None, val_transform=None, 
                channels_last=True, get_dataset=None,
                cache_train_loc=None, cache_val_loc=None):
        """
        :args:
        
        model: model.
        opt: optimizer.
        criterion: loss function. 
        FLAGS: (python dict) FLAGS containing information to pass in. Refer to FLAGS tutorial
             on what's required to put inside flag. 
        num_iter: (int) number of iterations. please use len(train_dataset) // batch_size
        train_ds: (IF num_iter IS NONE): (PyTorch Dataset) pass in training dataset. 
        momentum: Default: (0.95, 0.85)  # momentum for optimizer.
        div_factor: (int) Minimum learning rate: max_learning_rate / div_factor. 
        pct_start: (float) starting percentage. Defaults 0.4.
        # train_transform: Not used. 
        # val_transform: Not used. 
        channels_last: (bool) Whether to have NHWC instead of NCHW format. 
        get_dataset: (python function). Should returns train_ds, val_ds. Required if 
            you don't define cache_train_loc and cache_val_loc. 
        
        """
        self.SERIAL_EXEC = xmp.MpSerialExecutor()
        self.WRAPPED_MODEL = xmp.MpModelWrapper(model)
        self.opt = opt
        self.criterion = criterion
        self.flags = FLAGS
        self.chls = channels_last
        self.get_dataset = get_dataset
        self.cache_train_loc = cache_train_loc
        self.cache_val_loc = cache_val_loc
        
        if get_dataset is cache_train_loc is None:
            assert os.path.exists("./cache_train"), "Folder does not exist. Please put cached train dataset in this folder."
        if get_dataset is cache_val_loc is None:
            assert os.path.exists("./cache_val"), "Folder does not exist. Please put cached val dataset in this folder."
        
        if num_iter is train_ds is None: 
            raise ValueError("One of num_iter or train_ds must be defined")
        if num_iter is None: num_iter = len(train_ds) // self.flags["bs"]
        
        n = num_iter * self.flags["num_epochs"]
        a1 = int(n * pct_start)
        a2 = n - a1
        self.phases = ((a1, annealing_linear), (a2, annealing_cos))
        
        self.div_factor = div_factor
        self.momentum = momentum
        self.idx_s = 0
        
        max_lr = self.flags["lr"] * xm.xrt_world_size()
        min_lr = max_lr / self.div_factor
        self.lr_scheds = self.steps((min_lr, max_lr), (max_lr, min_lr / 1e4))
        self.mom_scheds = self.steps(self.momentum, self.momentum[::-1])
        self.update_lr_mom(self.lr_scheds[0].start, self.mom_scheds[0].start)

        # Future development
#         self.train_transform = train_transform
#         self.val_transform = val_transform
        
    def steps(self, *steps):
        """Build anneal schedule for all of the parameters. """
        return [Stepper(step, n_iter, func=func)
               for (step, (n_iter, func)) in zip(steps, self.phases)]
    
    def lrfinder(self, model, train_loader):
        if not xm.is_master_ordinal(): xm.rendezvous("lrfinder_once")
        lrfinder = LRFinder(model, self.opt, self.criterion, train_loader=train_loader)
        steepest, _, best_loss = lrfinder.lr_find()
        if xm.is_master_ordinal(): xm.rendezvous("lrfinder_once")
        
        max_lr = steepest * xm.xrt_world_size()
        min_lr = max_lr / self.div_factor
        self.lr_scheds = self.steps((min_lr, max_lr), (max_lr, min_lr / 1e4))
        self.mom_scheds = self.steps(self.momentum, self.momentum[::-1])
        self.update_lr_mom(self.lr_scheds[0].start, self.mom_scheds[0].start)
        
    def update_lr_mom(self, lr=0.001, mom=0.99):
        for l in self.opt.param_groups:
            l["lr"] = lr
            
            if isinstance(self.opt, (torch.optim.Adamax, torch.optim.Adam)):
                l["betas"] = (mom, 0.999)
            elif isinstance(self.opt, torch.optim.SGD):
                l["momentum"] = mom
    
    def train_tpu(self, train_ds, val_ds, fixed_lr=True, normalize=False, average=None):
        """
        :args fixed_lr: (bool) Whether to use a fixed_lr defined in FLAGS. 
            Defaults: True. 
        :args normalize: (bool/set of values) Whether to normalize the dataset 
            based on imagenet values. If passing in set of values, pass in as
            ((0.485, 0.456, 0.406), (0.229, 0.224, 0.225), 255) for example, 
            corresponding to (mean, std, max_pixel) respectively. 
        :args average: (str/None) f1_score arguments on average. 
        """
        torch.manual_seed(self.flags["seed"])
        
        if train_ds is None: train_ds, val_ds = self.SERIAL_EXEC.run(get_dataset)
        dls = dataloader(train_ds, val_ds, self.flags, distributed=True)
        
        device = xm.xla_device()
        model = self.WRAPPED_MODEL.to(device)
        
        
        if not fixed_lr: self.SERIAL_EXEC.run(lambda: self.lrfinder(model, dls["train"]))
        
#         def train_loop_fn(loader):
#             tracker = xm.RateTracker()
#             model.train()
            
#             running_loss = 0.0
#             total_samples = 0
            
#             for data, target in tqdm(loader):
#                 self.opt.zero_grad()
#                 data, target = data.to(device), target.to(device)
#                 if self.chls: data = data.to(memory_format=torch.channels_last)
                
#                 output = model(data)
#                 loss = self.criterion(output, target.to(torch.float32))
# #                 preds = (torch.sigmoid(output).data > 0.5).to(torch.float32)
                
#                 loss.backward()
#                 xm.optimizer_step(self.opt)
#                 self.update_lr_mom(self.lr_scheds[self.idx_s].step(),
#                                     self.mom_scheds[self.idx_s].step())
                            
#                 if self.lr_scheds[self.idx_s].is_done: self.idx_s += 1
#                 tracker.add(self.flags["bs"])
                
#                 running_loss += loss.item() * data.size(0)
#                 total_samples += data.size(0)

#             return running_loss, total_samples
            
#         def test_loop_fn(loader):
#             total_samples = 0
#             running_loss, f1Score = 0.0, 0.0
#             model.eval()
            
#             for data, target in tqdm(loader):
#                 data, target = data.to(device), target.to(device)
                
#                 output = model(data)
#                 loss = self.criterion(output, target.to(torch.float32))
#                 preds = (torch.sigmoid(output).data > 0.5).to(torch.float32)
                
#                 total_samples += data.size(0)
#                 running_loss += loss.item() * data.size(0)
                
#                 target = target.cpu().to(torch.int).numpy()
#                 preds = preds.cpu().to(torch.int).numpy()
                
#                 f1Score += f1_score(target, preds, average="weighted") * data.size(0)
                
#             epoch_loss = running_loss / total_samples
#             epoch_f1score = f1Score / total_samples
            
# #             print(f"""
# #                 Val loss: {epoch_loss} | 
# #                 Val F1Score: {epoch_f1score} | 
# #             """, flush=True)
#             return epoch_loss, epoch_f1score, data, preds, target

        mean, std = normalize_fn() if normalize else (None, None)
#         if mean is not None: mean, std = mean.to(device), std.to(device)
        
        for epoch in range(1, self.flags["num_epochs"] + 1):
            para_loader = pl.ParallelLoader(dls["train"], [device])
            running_loss, total_samples = self.train_loop_fn(
                                    para_loader.per_device_loader(device), model, mean, std)
            clear_output(wait=True)
            xm.master_print(f"Finished training epoch {epoch}")
            xm.master_print(f"Train loss: {running_loss / total_samples}", flush=True)
            
            para_loader = pl.ParallelLoader(dls["val"], [device])
            test_loss, f1score, data, pred, targ = self.test_loop_fn(
                                    para_loader.per_device_loader(device), model, average,
                                    mean, std)
            xm.master_print(f"""
                Val loss: {test_loss} | 
                Val F1Score: {f1score} | 
            """, flush=True)
            if self.flags["metrics_debug"]: xm.master_print(met.metrics_report(), flush=True)
                
        return test_loss, f1score, data, pred, targ, model
    
    def _mp_fn(self, rank, normalize):
        flags = self.flags
        torch.set_default_tensor_type(torch.FloatTensor)
        
        if self.get_dataset is None: 
            train_ds, val_ds = cached_dataset(self.cache_train_loc, self.cache_val_loc)
        else: train_ds, val_ds = None, None
            
        loss, f1score, data, pred, targ, model = self.train_tpu(train_ds, val_ds, normalize=normalize)
        if rank == 0: torch.save(model.state_dict(), self.flags["save_path"])
            
    def train(self, normalize=False):
        import gc
        gc.collect()
        xmp.spawn(self._mp_fn, args=(normalize, ), nprocs=self.flags["num_cores"], start_method="fork")

    def train_loop_fn(self, loader, model, mean, std):
        tracker = xm.RateTracker()
        model.train()
        
        running_loss = 0.0
        total_samples = 0
        
        # get one batch and put onto same device. Alternatively can put in loop. 
        for data, target in loader: break  
        if mean is not None: mean, std = mean.to(data.device), std.to(data.device)
        
        for data, target in tqdm(loader):
            self.opt.zero_grad()
#             data, target = data.to(device), target.to(device)
            if mean is not None: data = (data - mean) / std
            if self.chls: data = data.to(memory_format=torch.channels_last)
            
            output = model(data)
            loss = self.criterion(output, target.to(torch.float32))
#                 preds = (torch.sigmoid(output).data > 0.5).to(torch.float32)
            
            loss.backward()
            xm.optimizer_step(self.opt)
            self.update_lr_mom(self.lr_scheds[self.idx_s].step(),
                                self.mom_scheds[self.idx_s].step())
                        
            if self.lr_scheds[self.idx_s].is_done: self.idx_s += 1
            tracker.add(self.flags["bs"])
            
            running_loss += loss.item() * data.size(0)
            total_samples += data.size(0)

        return running_loss, total_samples

    def test_loop_fn(self, loader, model, average, mean, std):
        total_samples = 0
        running_loss, f1Score = 0.0, 0.0
        model.eval()
        
        # get one batch and put onto same device. Alternatively can put in loop. 
        for data, target in loader: break
        if mean is not None: mean, std = mean.to(data.device), std.to(data.device)
        
        for data, target in tqdm(loader):
#             data, target = data.to(device), target.to(device)
            if mean is not None: data = (data - mean) / std
            if self.chls: data = data.to(memory_format=torch.channels_last)
            
            output = model(data)
            loss = self.criterion(output, target.to(torch.float32))
            preds = (torch.sigmoid(output).data > 0.5).to(torch.float32)
            
            total_samples += data.size(0)
            running_loss += loss.item() * data.size(0)
            
            target = target.cpu().to(torch.int).numpy()
            preds = preds.cpu().to(torch.int).numpy()
            
            f1Score += f1_score(target, preds, average=average) * data.size(0)
            
        epoch_loss = running_loss / total_samples
        epoch_f1score = f1Score / total_samples
        
#             print(f"""
#                 Val loss: {epoch_loss} | 
#                 Val F1Score: {epoch_f1score} | 
#             """, flush=True)
        return epoch_loss, epoch_f1score, data, preds, target

# %% [markdown]
# `weighted` can result in F-score **that is not between precision and recall**. URL: https://scikit-learn.org/stable/modules/generated/sklearn.metrics.f1_score.html

# %% [markdown]
# # Predict Batch
# For batch prediction

# %% [code] {"jupyter":{"outputs_hidden":false}}
def predict_batch(model, dl_test, paths=None, normalize=False, device=None):
    """
    Batch prediction. 
    File predicted will be according to how the data is input. It should be in the order 
    of your input paths. 
    :args: 
        model (PyTorch model): state dict should already be loaded by yourself. 
        dl_test: (PyTorch Dataloaders) Dataloaders containing test files. 
        paths (Python List of PosixPaths/str): paths to test files. (currently not used)
        normalize (bool/tuple): Whether to normalize data or not. If you want
            True, please pass in tuple of mean and std for non-Imagenet normalization. 
    :return: (Tensor) list of predictions in the order of paths. 
    """
    with torch.no_grad():
        model = model.to(device)
        model.eval()
        if normalize == True: mean, std = normalize_fn()
        elif type(normalize) == tuple: 
            mean, std = normalize
            normalize = True
            
        total = None
        
        for k, data in enumerate(tqdm(dl_test)):
            if normalize: data = (data - mean) / std
            output = model(data.to(device))
            if total is None: total = copy.deepcopy(output.cpu())
            else: total = torch.cat([total, output.cpu()], axis=0)
            
    return total

# CHanges with multiprocessing: if the class to fetch data can return the idx, then we can 
# use it to search back the filename based on the idx. Idx as integers can be collated together 
# unlike file name whom are str/object. 



def setup_colab():
    os.system("pip install cloud-tpu-client==0.10 https://storage.googleapis.com/tpu-pytorch/wheels/torch_xla-1.9-cp37-cp37m-linux_x86_64.whl")
    print("Download complete.")
    clear_output()


import os
from IPython.display import clear_output

try: import torch_xla
except Exception: setup_colab()

from .tpu_utility_1 import *
from .tpu_cache_ds_utils import *
from .other_utils import *
from ._lr_finder import *